/* this file is CLEAN2 [sverrehu] */

/* Procedures dealing with grey-scale and mono dithering, as well as X
 * Windows set up procedures. */

/* 
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/* 
 * Portions of this software Copyright (c) 1995 Brown University.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL BROWN UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF BROWN
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * BROWN UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
 * BASIS, AND BROWN UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <unistd.h> /* select */

#include "video.h"
#include "proto.h"
#include "dither.h"
#include "mpeg_proc.h"
#include "xtra.h"

/* Array that remaps color numbers to actual pixel values used by X
 * server. */
unsigned char pixel[256];
unsigned long wpixel[256];

/* Arrays holding quantized value ranged for lum, cr, and cb. */
int *lum_values;
int *cr_values;
int *cb_values;

/* Luminance and chrominance lookup tables */
static double *L_tab, *Cr_r_tab, *Cr_g_tab, *Cb_g_tab, *Cb_b_tab;

/* 
 *--------------------------------------------------------------
 *
 * InitColor --
 *
 *      Initialize lum, cr, and cb quantized range value arrays.
 *      Also initializes the lookup tables for the possible
 *      values of lum, cr, and cb.
 *    Color values from ITU-R BT.470-2 System B, G and SMPTE 170M
 *    see InitColorDither in 16bits.c for more
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitColor(void)
{
    int i;

    L_tab = (double *) malloc(LUM_RANGE * sizeof(double));
    Cr_r_tab = (double *) malloc(CR_RANGE * sizeof(double));
    Cr_g_tab = (double *) malloc(CR_RANGE * sizeof(double));
    Cb_g_tab = (double *) malloc(CB_RANGE * sizeof(double));
    Cb_b_tab = (double *) malloc(CB_RANGE * sizeof(double));

    if (L_tab == NULL || Cr_r_tab == NULL || Cr_g_tab == NULL
	|| Cb_g_tab == NULL || Cb_b_tab == NULL) {
	fprintf(stderr, "out of memory\n");
	exit(1);
    }
    for (i = 0; i < LUM_RANGE; i++) {
	lum_values[i] = ((i * 256) / (LUM_RANGE)) + (256 / (LUM_RANGE * 2));
	L_tab[i] = lum_values[i];
	if (gammaCorrectFlag) {
	    L_tab[i] = GAMMA_CORRECTION(L_tab[i]);
	}
    }

    for (i = 0; i < CR_RANGE; i++) {
	register double tmp;

	if (chromaCorrectFlag) {
	    tmp = ((i * 256) / (CR_RANGE)) + (256 / (CR_RANGE * 2));
	    Cr_r_tab[i] = (int) (0.419 / 0.299)
		* CHROMA_CORRECTION128D(tmp - 128.0);
	    Cr_g_tab[i] = (int) -(0.299 / 0.419)
		* CHROMA_CORRECTION128D(tmp - 128.0);
	    cr_values[i] = CHROMA_CORRECTION256(tmp);
	} else {
	    tmp = ((i * 256) / (CR_RANGE)) + (256 / (CR_RANGE * 2));
	    Cr_r_tab[i] = (int) (0.419 / 0.299) * (tmp - 128.0);
	    Cr_g_tab[i] = (int) -(0.299 / 0.419) * (tmp - 128.0);
	    cr_values[i] = (int) tmp;
	}
    }

    for (i = 0; i < CB_RANGE; i++) {
	register double tmp;
	if (chromaCorrectFlag) {
	    tmp = ((i * 256) / (CB_RANGE)) + (256 / (CB_RANGE * 2));
	    Cb_g_tab[i] = (int) -(0.114 / 0.331)
		* CHROMA_CORRECTION128D(tmp - 128.0);
	    Cb_b_tab[i] = (int) (0.587 / 0.331)
		* CHROMA_CORRECTION128D(tmp - 128.0);
	    cb_values[i] = CHROMA_CORRECTION256(tmp);
	} else {
	    tmp = ((i * 256) / (CB_RANGE)) + (256 / (CB_RANGE * 2));
	    Cb_g_tab[i] = (int) -(0.114 / 0.331) * (tmp - 128.0);
	    Cb_b_tab[i] = (int) (0.587 / 0.331) * (tmp - 128.0);
	    cb_values[i] = (int) tmp;
	}
    }
}

/* 
 *--------------------------------------------------------------
 *
 * ConvertColor --
 *
 *      Given a l, cr, cb tuple, converts it to r,g,b.
 *
 * Results:
 *      r,g,b values returned in pointers passed as parameters.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

static void ConvertColor(unsigned int l, unsigned int cr, unsigned int cb,
			 unsigned char *r, unsigned char *g, unsigned char *b)
{
    double fl, fr, fg, fb;

    fl = L_tab[l];
    fr = fl + Cr_r_tab[cr];
    fg = fl + Cr_g_tab[cr] + Cb_g_tab[cb];
    fb = fl + Cb_b_tab[cb];

    if (fr < 0.0)
	fr = 0.0;
    else if (fr > 255.0)
	fr = 255.0;

    if (fg < 0.0)
	fg = 0.0;
    else if (fg > 255.0)
	fg = 255.0;

    if (fb < 0.0)
	fb = 0.0;
    else if (fb > 255.0)
	fb = 255.0;

    *r = (unsigned char) fr;
    *g = (unsigned char) fg;
    *b = (unsigned char) fb;
}

#ifdef SH_MEM
int gXErrorFlag = 0;

static int HandleXError(Display *dpy, XErrorEvent *event)
{
    gXErrorFlag = 1;
    return 0;
}

void InstallXErrorHandler(Display *display)
{
    XSetErrorHandler(HandleXError);
    XFlush(display);
}

void DeInstallXErrorHandler(Display *display)
{
    XSetErrorHandler(NULL);
    XFlush(display);
}
#endif /* ifdef SH_MEM */

/* 
 *--------------------------------------------------------------
 *
 * ResizeDisplay --
 *
 *      Resizes display window.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void ResizeDisplay(unsigned int w, unsigned int h, XInfo *xinfo)
{
    XResizeWindow(xinfo->display, xinfo->window, w, h);
    XFlush(xinfo->display);
}

static Bool isMapNotifyEvent(Display *display, XEvent *xev, char *data)
{
    XInfo *xinfo = (XInfo *) data;

    return xev->type == MapNotify && xev->xmap.event == xinfo->window;
}

/* 
 *--------------------------------------------------------------
 *
 * MakeWindow --
 *
 *      Create X Window
 *
 * Results:
 *      Read the code.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

#ifdef SH_MEM
static int CompletionType = -1;
#endif

static int MakeWindow(char *name, XInfo *xinfo)
{
    Display *display;
    int screen;
    Visual *vis;
    XSetWindowAttributes attrib;
    unsigned long attrib_flags = 0;
    XVisualInfo vinfo;
    XEvent xev;

    if (xinfo == NULL)
	return 0;
    if (xinfo->display == NULL) {
	display = xinfo->display = XOpenDisplay(name);
	if (xinfo->display == NULL) {
	    fprintf(stderr, "cannot open display\n");
	    exit(-2);
	}
    }
    display = xinfo->display;
#ifdef SH_MEM
    if (shmemFlag)
	CompletionType = XShmGetEventBase(display) + ShmCompletion;
#endif
    screen = DefaultScreen(display);
    /* Fill in hint structure */
    if (xinfo->hints.width == 0) {
	xinfo->hints.width = 150;
	xinfo->hints.height = 150;
    }
    xinfo->hints.flags = PPosition | PSize;
    /* Make the window */
    if (xinfo->ditherType == FULL_COLOR_DITHER) {
	CreateFullColorWindow(xinfo);
	if (xinfo->window == 0) {
	    fprintf(stderr,
		    "unable to do full color dithering on this display\n");
	    exit(1);
	}
    }
    if (xinfo->ditherType != FULL_COLOR_DITHER) {
	if (!XMatchVisualInfo(display, screen, xinfo->depth = 8, PseudoColor,
			      &vinfo)) {
	    if (xinfo->ditherType != GRAY_DITHER &&
		xinfo->ditherType != GRAY2_DITHER) {
		fprintf(stderr, "requires at least 8 bit display\n");
		exit(1);
	    } else if (!XMatchVisualInfo(display, screen, xinfo->depth = 32,
					 GrayScale, &vinfo) &&
		       !XMatchVisualInfo(display, screen, xinfo->depth = 24,
					 GrayScale, &vinfo) &&
		       !XMatchVisualInfo(display, screen, xinfo->depth = 16,
					 GrayScale, &vinfo) &&
		       !XMatchVisualInfo(display, screen, xinfo->depth = 8,
					 GrayScale, &vinfo) &&
		       !XMatchVisualInfo(display, screen, xinfo->depth = 32,
					 TrueColor, &vinfo) &&
		       !XMatchVisualInfo(display, screen, xinfo->depth = 24,
					 TrueColor, &vinfo) &&
		       !XMatchVisualInfo(display, screen, xinfo->depth = 16,
					 TrueColor, &vinfo)) {
		fprintf(stderr, "requires at least 8 bit display\n");
		exit(1);
	    }
	}
	vis = vinfo.visual;
	if (XDefaultDepthOfScreen(XDefaultScreenOfDisplay(display)) != 8
	    || xinfo->owncmFlag) {
	    attrib_flags |= CWColormap;
	    attrib.colormap = XCreateColormap(display, xinfo->parent,
					      vis, AllocNone);
	    xinfo->owncmFlag = 1;
	}
	attrib.background_pixel = WhitePixel(display, screen);;
	attrib.border_pixel = BlackPixel(display, screen);
	attrib.backing_store = NotUseful;
	attrib.save_under = False;
	attrib_flags |= CWBackPixel | CWBorderPixel
	    | CWBackingStore | CWSaveUnder;
	xinfo->window = XCreateWindow(display, xinfo->parent,
				      xinfo->hints.x, xinfo->hints.y,
				      xinfo->hints.width, xinfo->hints.height,
				      0, xinfo->depth, InputOutput, vis,
				      attrib_flags, &attrib);
    }
    XSelectInput(display, xinfo->window, StructureNotifyMask);
    /* Map window. */
    XMapWindow(display, xinfo->window);
    /* Wait for map. */
    XIfEvent(display, &xev, isMapNotifyEvent, (char *) xinfo);
    XSelectInput(display, xinfo->window, NoEventMask);
    return TRUE;
}

/* 
 *--------------------------------------------------------------
 *
 * InitDisplay --
 *
 *      Initialized display, sets up colormap, etc.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitDisplay(char *name, XInfo *xinfo)
{
    int ncolors = LUM_RANGE * CB_RANGE * CR_RANGE;
    XColor xcolor;
    int i, lum_num, cr_num, cb_num;
    unsigned char r, g, b;
    Colormap dcmap;
    Display *display;

    if (xinfo == NULL)
	return;
    if (!MakeWindow(name, xinfo)) {
	/* Could not do that dither.  Try again if can */
	switch (xinfo->ditherType) {
	  case ORDERED2_DITHER:
	  case HYBRID_DITHER:
	  case FS2FAST_DITHER:
	  case Twox2_DITHER:
	      xinfo->ditherType = FULL_COLOR_DITHER;
	      InitColorDisplay(name, xinfo);
	      InitColorDither(xinfo->depth == 32);
	      return;
	  case FULL_COLOR_DITHER:
	  case GRAY_DITHER:
	  case GRAY2_DITHER:
	  default:
	      /* cant do anything */
	      exit(-1);
	}
    }
    if (xinfo != NULL) {
	display = xinfo->display;
	xinfo->gc = XCreateGC(display, xinfo->window, 0, 0);
	dcmap = xinfo->cmap
	    = xGetParentColormap(display, xinfo->window);
	xcolor.flags = DoRed | DoGreen | DoBlue;
	if (xinfo->owncmFlag)
	    goto create_map;
    }
  retry_alloc_colors:
    for (i = 0; i < ncolors; i++) {
	lum_num = (i / (CR_RANGE * CB_RANGE)) % LUM_RANGE;
	cr_num = (i / CB_RANGE) % CR_RANGE;
	cb_num = i % CB_RANGE;
	ConvertColor(lum_num, cr_num, cb_num, &r, &g, &b);
	xcolor.red = r * 256;
	xcolor.green = g * 256;
	xcolor.blue = b * 256;
	if (xinfo != NULL
	    && xAllocClosestColor(display, xinfo->cmap, &xcolor) == 0) {
	    if (xinfo->cmap == dcmap) {
		int j;
		unsigned long tmp_pixel;
		XWindowAttributes xwa;

		/* Free colors. */
		for (j = 0; j < i; j++) {
		    tmp_pixel = wpixel[j];
		    XFreeColors(display, xinfo->cmap, &tmp_pixel, 1, 0);
		}
	    create_map:
                XGetWindowAttributes(display, xinfo->window, &xwa);
		xinfo->cmap = XCreateColormap(display, xinfo->window,
					      xwa.visual, AllocNone);
		XSetWindowColormap(display, xinfo->window, xinfo->cmap);
		xinfo->owncmFlag = 1;
		goto retry_alloc_colors;
	    }
	}
	pixel[i] = xcolor.pixel;
	wpixel[i] = xcolor.pixel;
    }
    xinfo->ximage = NULL;
    if (xinfo->applicationMainWindow && xinfo->owncmFlag)
	xAddWMColormapWindow(display, xinfo->applicationMainWindow,
			     xinfo->window);
}

/* 
 *--------------------------------------------------------------
 *
 * InitGrayDisplay --
 *
 *      Initialized display for gray scale dither.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitGrayDisplay(char *name, XInfo *xinfo)
{
    int ncolors = 128;
    XColor xcolor;
    int i;
    Colormap dcmap;
    unsigned long tmp_pixels[256];
    Window window;
    Display *display;

    MakeWindow(name, xinfo);
    display = xinfo->display;
    window = xinfo->window;
    xinfo->gc = XCreateGC(display, window, 0, 0);
    dcmap = xinfo->cmap = xGetParentColormap(display, xinfo->window);
    xcolor.flags = DoRed | DoGreen | DoBlue;
    if (xinfo->owncmFlag)
	goto create_map;
  retry_alloc_grays:
    for (i = 0; i < ncolors; i++) {
	xcolor.red = xcolor.green = xcolor.blue
	    = GAMMA_CORRECTION(i * 2) * 256;
	if (xAllocClosestColor(display, xinfo->cmap, &xcolor) == 0
	    && xinfo->cmap == dcmap) {
	    int j;
	    XWindowAttributes xwa;

	    /* Free colors. */
	    for (j = 0; j < i; j++) {
		unsigned long tmp_pixel;
		tmp_pixel = tmp_pixels[j * 2];
		XFreeColors(display, xinfo->cmap, &tmp_pixel, 1, 0);
	    }
	  create_map:
	    XGetWindowAttributes(display, window, &xwa);
	    xinfo->cmap = XCreateColormap(display, window,
					  xwa.visual, AllocNone);
	    XSetWindowColormap(display, window, xinfo->cmap);
	    xinfo->owncmFlag = 1;
	    goto retry_alloc_grays;
	}
	tmp_pixels[i * 2] = pixel[i * 2] = xcolor.pixel;
	tmp_pixels[(i * 2) + 1] = pixel[(i * 2) + 1] = xcolor.pixel;
	wpixel[(i * 2)] = xcolor.pixel;
	wpixel[(i * 2) + 1] = xcolor.pixel;
	if (xinfo->depth == 8) {
	    wpixel[i * 2] |= wpixel[i * 2] << 8;
	    wpixel[i * 2 + 1] |= wpixel[i * 2 + 1] << 8;
	}
	if (xinfo->depth == 8 || xinfo->depth == 16) {
	    wpixel[i * 2] |= wpixel[i * 2] << 16;
	    wpixel[i * 2 + 1] |= wpixel[i * 2 + 1] << 16;
	}
    }
    xinfo->ximage = NULL;
    if (xinfo->applicationMainWindow && xinfo->owncmFlag)
	xAddWMColormapWindow(display, xinfo->applicationMainWindow,
			     xinfo->window);
}

/* 
 *--------------------------------------------------------------
 *
 * InitColorDisplay --
 *
 *      Initialized display for full color output.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitColorDisplay(char *name, XInfo *xinfo)
{
    XWindowAttributes winattr;

    if (xinfo == NULL)
	return;
    MakeWindow(name, xinfo);
    if (xinfo->gc == 0)
	xinfo->gc = XCreateGC(xinfo->display, xinfo->window, 0, 0);
    xinfo->ximage = NULL;
    XGetWindowAttributes(xinfo->display, xinfo->window, &winattr);
    /* Misuse of wpixel */
    wpixel[0] = winattr.visual->red_mask;
    wpixel[1] = winattr.visual->green_mask;
    wpixel[2] = winattr.visual->blue_mask;
}

static Bool isCompletionEvent(Display *display, XEvent *xev, char *data)
{
    return xev->type == CompletionType;
}

/* 
 *--------------------------------------------------------------
 *
 * ExecuteDisplay --
 *
 *      Actually displays display plane in previously created window.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void ExecuteDisplay(VidStream *vid_stream, XInfo *xinfo)
{
    char dummy;
    Visual *fc_visual;
    int depth;
#ifdef ED_TIMING
    static int rate_deal = -1;
    static int one_frame_time;
    static struct timeval tftarget, tfnow;
    register int usec, sec;
#endif
    Display *display = NULL;
    XEvent xev;

    if (xinfo != NULL)
	display = xinfo->display;
    vid_stream->totNumFrames++;
#ifdef ED_TIMING
    /* Do frame rate control */
    switch (rate_deal) {
      case 0:
	break;
      default:
	gettimeofday(&tfnow, (struct timezone *) NULL);
	usec = tftarget.tv_usec - tfnow.tv_usec;
	sec = tftarget.tv_sec - tfnow.tv_sec;
	if (usec < 0) {
	    usec += 1000000;
	    sec--;
	}
	/* If we're not behind, wait a bit */
	if ((sec >= 0) && usec > 0) {
	    tfnow.tv_sec = sec;
	    tfnow.tv_usec = usec;
	    select(0, NULL, NULL, NULL, &tfnow);
	    gettimeofday(&tfnow, (struct timezone *) NULL);
	}
	/* Setup target for next frame */
	tftarget.tv_usec = tfnow.tv_usec + one_frame_time;
	if (tftarget.tv_usec >= 1000000) {
	    tftarget.tv_usec -= 1000000;
	    tftarget.tv_sec = tfnow.tv_sec + 1;
	} else
	    tftarget.tv_sec = tfnow.tv_sec;
	break;
      case -1:
	switch (framerate) {
	  case -1:		/* Go with stream Value */
	    rate_deal = vid_stream->fps;
	    gettimeofday(&tftarget, (struct timezone *) NULL);
	    one_frame_time = 1000000 / rate_deal;
	    break;
	  case 0:		/* as fast as possible */
	    rate_deal = 0;
	    break;
	  default:
	    rate_deal = framerate;
	    gettimeofday(&tftarget, (struct timezone *) NULL);
	    one_frame_time = 1000000 / rate_deal;
	    break;
	}
	break;
    }
#endif
    if (xinfo == NULL)
	return;
    if (xinfo->ximage == NULL) {
	int pixels_per_mb = 16, w, h;
	
	if (IS_2x2_DITHER(xinfo->ditherType))
	    pixels_per_mb = 32;
	w = vid_stream->mb_width * pixels_per_mb;
	h = vid_stream->mb_height * pixels_per_mb;
	if (xinfo->ditherType == FULL_COLOR_DITHER) {
	    fc_visual = FindFullColorVisual(display, &depth);
	    xinfo->ximage = XCreateImage(display, fc_visual, depth,
					 ZPixmap, 0, &dummy, w, h, 32, 0);
	} else
	    xinfo->ximage
		= XCreateImage(display, None, xinfo->depth,
			       ZPixmap, 0, &dummy, w, h, 8, 0);
	mpDoMove(w, h);
    }
/* Always work in native bit and byte order. This tells Xlib to
 * reverse bit and byte order if necessary when crossing a
 * network. Frankly, this part of XImages is somewhat underdocumented,
 * so this may not be exactly correct.  */
#ifdef LITTLE_ENDIAN_ARCHITECTURE
    xinfo->ximage->byte_order = LSBFirst;
    xinfo->ximage->bitmap_bit_order = LSBFirst;
#else
    xinfo->ximage->byte_order = MSBFirst;
    xinfo->ximage->bitmap_bit_order = MSBFirst;
#endif
#ifdef SH_MEM
    if (shmemFlag) {
	XShmPutImage(display, xinfo->window, xinfo->gc,
		     vid_stream->current->ximage, 0, 0, 0, 0,
		     vid_stream->current->ximage->width,
		     vid_stream->current->ximage->height, True);
	XFlush(display);
	XIfEvent(display, &xev, isCompletionEvent, NULL);
    } else
#endif /* ifdef SH_MEM */
	{
	    xinfo->ximage->data = (char *) vid_stream->current->display;
	    XPutImage(display, xinfo->window, xinfo->gc, xinfo->ximage,
		      0, 0, 0, 0,
		      xinfo->ximage->width, xinfo->ximage->height);
	}
}
