/* $Id$ */
/**************************************************************************
 *
 *  FILE            mpeg_api.c
 *  MODULE OF       mpeg_play library.
 *
 *  DESCRIPTION     The library API.
 *
 *  WRITTEN BY      Sverre H. Huseby <sverrehu@ifi.uio.no>
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "mpeg_proc.h"
#include "mpeg_play.h"

#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif

/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

char *
mpRetStr(int n)
{
    switch (n) {
      case MPEG_ERR_CHILD:
	return "no reply from video decoding process.";
      case MPEG_ERR_UNKNOWN:
	return "video decoding process got an unknown command. "
	       "this should never happen.";
      case MPEG_ERR_OVERFLOW:
	return "command buffer overflow in video decoding process.";
      case MPEG_ERR_NO_MOVIE:
	return "no movie present.";
      case MPEG_ERR_NOT_SEEKABLE:
	return "stream not seekable.";
      case MPEG_OK:
	return "ok.";
      case MPEG_FILM_HAS_ENDED:
	return "end of movie reached.";
    }
    return "???";
}

MpegStream
mpNewStreamFromFile(const char *filename,
		    Display *display,
		    Window parent,
		    Window applicationMainWindow,
		    DitherType dither,
		    int frameRate,
		    int loop,
		    double gammaCorrection,
		    double chromaCorrection,
		    int useSharedMemory,
		    int useOwnColorMap)
{
    MpegStream ret;

    ret = mpNewStream(&display);  /* display changed in child */
    if (!mpThisIsTheChild)
	return ret;

    mpChildNewStreamFromFile(filename, display, parent,
			     applicationMainWindow,
			     dither, frameRate, loop,
			     gammaCorrection, chromaCorrection,
			     useSharedMemory, useOwnColorMap);

    return NULL; /* never gets here, actually - just please the compiler. */
}

MpegStream
mpNewStreamPushed(Display *display,
		  Window parent,
		  Window applicationMainWindow,
		  DitherType dither,
		  int frameRate,
		  double gammaCorrection,
		  double chromaCorrection,
		  int useSharedMemory,
		  int useOwnColorMap)
{
    MpegStream ret;

    ret = mpNewStream(&display);  /* display changed in child */
    if (!mpThisIsTheChild)
	return ret;

    mpChildNewStreamPushed(display, parent,
			   applicationMainWindow,
			   dither, frameRate,
			   gammaCorrection, chromaCorrection,
			   useSharedMemory, useOwnColorMap);

    return NULL; /* never gets here, actually - just please the compiler. */
}

int
mpPlay(MpegStream stream)
{
    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    return mpServerSendCommand(stream, MPEG_CMD_PLAY);
}

int
mpNextFrame(MpegStream stream)
{
    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    return mpServerSendCommand(stream, MPEG_CMD_NEXT_FRAME);
}

int
mpRewind(MpegStream stream)
{
    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    return mpServerSendCommand(stream, MPEG_CMD_REWIND);
}

int
mpStop(MpegStream stream)
{
    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    return mpServerSendCommand(stream, MPEG_CMD_PAUSE);
}

int
mpQuit(MpegStream stream)
{
    int ret;

    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    ret = mpServerSendCommand(stream, MPEG_CMD_QUIT);
    waitpid(stream->child, NULL, 0);
    return ret;
}

int
mpRepaint(MpegStream stream)
{
    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    return mpServerSendCommand(stream, MPEG_CMD_REPAINT);
}

int
mpParentResize(MpegStream stream)
{
    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    return mpServerSendCommand(stream, MPEG_CMD_PARENT_RESIZE);
}

int
mpSetLoop(MpegStream stream, int onoff)
{
    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    return mpServerSendIntCommand(stream, MPEG_CMD_SET_LOOP, onoff);
}

int
mpQueryWantedBytes(MpegStream stream, int *n)
{
    int ret;

    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    if ((ret = mpServerSendCommand(stream,
				   MPEG_CMD_QUERY_WANTED_BYTES)) != MPEG_OK)
	return ret;
    *n = mpServerReceiveInt(stream);
    return ret;
}

int
mpSendBytes(MpegStream stream, void *buffer, int *n)
{
#ifdef PIPE_BUF
#define BLOCK_SIZE PIPE_BUF
#else
#define BLOCK_SIZE 512
#endif
    int ret, q;

    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    if ((ret = mpServerSendIntCommand(stream, MPEG_CMD_SEND_BYTES,
				      *n)) != MPEG_OK)
	return ret;
    if ((ret = mpServerReceiveInt(stream)) < -1)
	return ret;
    *n = ret;
    for (q = 0; q < *n; q += BLOCK_SIZE)
	if (write(stream->pipeCmd, buffer + q, min(*n - q, BLOCK_SIZE)) < 0) {
	    perror("mpSendBytes");
	    return MPEG_ERR_CHILD;
	}
	
    return mpServerReceiveStatus(stream);
}

int
mpEndOfStream(MpegStream stream)
{
    if (!stream)
	return MPEG_ERR_NO_MOVIE;
    return mpServerSendCommand(stream, MPEG_CMD_END_OF_STREAM);
}
