/* $Id$ */
/**************************************************************************
 *
 *  FILE            mpeg_proc.c
 *  MODULE OF       mpeg_play library.
 *
 *  DESCRIPTION     
 *
 *  WRITTEN BY      Sverre H. Huseby <sverrehu@ifi.uio.no>
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>

#include "video.h"
#include "proto.h"
#include "util.h"
#include "dither.h"
#include "xtra.h"
#include "xfile.h"
#include "cmdq.h"
#include "xselect.h"
#include "mpeg_proc.h"

#define BUF_LENGTH_FILE   80000
#define BUF_LENGTH_PUSHED 4096
#define BUF_LENGTH_XFILE  16384

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

static int pipeCmd;  /* read end of the command pipe */
static int pipeStat; /* write end of the status pipe */

static XInfo xinfo;
static VidStream *stream;
static int endOfStream;

/**************************************************************************
 *                                                                        *
 *                        P U B L I C    D A T A                          *
 *                                                                        *
 **************************************************************************/

/* Loop flag. */
int loopFlag = 0;

/* Shared memory flag. */
int shmemFlag = 0;

/* Framerate, -1: specified in stream (default) 0: as fast as possible N
 * (N>0): N frames/sec   */
int framerate = -1;

/* Flag for gamma correction */
int gammaCorrectFlag = 0;
double gammaCorrect = 1.0;

/* Flag for chroma correction */
int chromaCorrectFlag = 0;
double chromaCorrect = 1.0;

int mpThisIsTheChild = 0;

/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static int
mpStreamIsSeekable(void)
{
    return (stream->input != NULL);
}

static void
mpDoRepaint(void)
{
    if (stream == NULL || stream->current == NULL || xinfo.ximage == NULL)
	return;

#ifdef SH_MEM
    if (shmemFlag) {
	XShmPutImage(xinfo.display, xinfo.window, xinfo.gc,
		     stream->current->ximage, 0, 0, 0, 0,
		     stream->current->ximage->width,
		     stream->current->ximage->height, 1);
	XFlush(xinfo.display);
    } else
#endif /* ifdef SH_MEM */
	{
	    xinfo.ximage->data = (char *) stream->current->display;
	    XPutImage(xinfo.display, xinfo.window, xinfo.gc,
		      xinfo.ximage, 0, 0, 0, 0,
		      xinfo.ximage->width, xinfo.ximage->height);
	}
}

static int
mpCommonInit(Display *display,
	     Window parent,
	     Window applicationMainWindow,
	     DitherType *dither,
	     int frameRate,
	     int loop,
	     double gamma,
	     double chroma,
	     int useSharedMemory,
	     int useOwnColorMap)
{
    if (gamma <= 0.0)
	gamma = 1.0;
    gammaCorrect = gamma;
    gammaCorrectFlag = (gamma != 1.0);
    if (chroma <= 0.0)
	chroma = 1.0;
    chromaCorrect = chroma;
    chromaCorrectFlag = (chroma != 1.0);
    framerate = frameRate;
    loopFlag = loop;

#ifdef SH_MEM
    shmemFlag = useSharedMemory;
#endif

    lum_values = (int *) malloc(LUM_RANGE * sizeof(int));
    cr_values = (int *) malloc(CR_RANGE * sizeof(int));
    cb_values = (int *) malloc(CB_RANGE * sizeof(int));

    init_tables(); /* no memory allocated in this one */

    xinfo.display = display;
    xinfo.parent = parent;
    xinfo.applicationMainWindow = applicationMainWindow;
    /* xinfo.ximage is set to null later */
    xinfo.ditherType = *dither;
    xinfo.owncmFlag = useOwnColorMap;
    xinfo.hints.x = 0;
    xinfo.hints.y = 0;
    xinfo.hints.width = 160;
    xinfo.hints.height = 100;
    xinfo.visual = NULL;
    xinfo.cmap = 0;
    xinfo.gc = 0;

    switch (xinfo.ditherType) {
      case ORDERED2_DITHER:
	InitColor();
	InitDisplay("", &xinfo);
	InitOrdered2Dither();
	break;

      case HYBRID_DITHER:
	InitColor();
	InitHybridDither();
	InitDisplay("", &xinfo);
	break;

      case FS2FAST_DITHER:
	InitColor();
	InitFS2FastDither();
	InitDisplay("", &xinfo);
	break;

      case Twox2_DITHER:
	InitColor();
	Init2x2Dither();
	InitDisplay("", &xinfo);
	PostInit2x2Dither();
	break;

      case FULL_COLOR_DITHER:
	InitColorDisplay("", &xinfo);
	InitColorDither(xinfo.depth >= 24);
	break;

      case GRAY_DITHER:
      case GRAY2_DITHER:
	InitGrayDisplay("", &xinfo);
	break;
    }
    *dither = xinfo.ditherType;

#ifdef SH_MEM
    if (shmemFlag && xinfo.display != NULL)
	if (!XShmQueryExtension(xinfo.display))
	    shmemFlag = 0;
#endif

    InitCrop();
    mpDoMove(0, 0);

    return 0;
}

static int
mpCommonFinish(void)
{
    xRemoveWMColormapWindow(xinfo.display,
			    xinfo.applicationMainWindow, xinfo.window);
    XDestroyWindow(xinfo.display, xinfo.window);
    DestroyVidStream(stream, &xinfo);
    free(lum_values);
    free(cr_values);
    free(cb_values);
    lum_values = cr_values = cb_values = NULL;

    return 0;
}

static void
mpDoRewind(void)
{
    if (!mpStreamIsSeekable())
	return;
    rewind(stream->input);
    clear_data_stream(stream);
    ResetVidStream(stream);
    mpegVidRsrc(0, stream, 1, &xinfo);
    while (stream->totNumFrames == 0)
	mpegVidRsrc(0, stream, 0, &xinfo);
}

static void
mpDoAdvanceFrame(void)
{
    int frame;

    frame = stream->totNumFrames;
    do {
	mpegVidRsrc(0, stream, 0, &xinfo);
    } while (stream->totNumFrames == frame);
}

static void
mpChildLoop(void)
{
    volatile int firstTime = 1, playing = 0;
    int cmd, param;
    int doNextFrame;
    struct timeval tv, perImage, used, before, after;

    for (;;) {
	if (firstTime) {
	    int mult;

	    firstTime = 0;
	    endOfStream = 0;
	    tv.tv_sec = tv.tv_usec = 0;
	    stream->ditherType = xinfo.ditherType;
	    stream->matched_depth = xinfo.depth;
	    if (mpegVidRsrc(0, stream, 1, &xinfo) == NULL) {
		XDestroyWindow(xinfo.display, xinfo.window);
		/* stream has already been destroyed */
		fprintf(stderr, "not an MPEG-stream\n");
		exit(1);
	    }
	    if (IS_2x2_DITHER(xinfo.ditherType))
		mult = 2;
	    else
		mult = 1;
	    if (stream != NULL)
		ResizeDisplay((unsigned int) stream->h_size * mult,
			      (unsigned int) stream->v_size * mult,
			      &xinfo);
	    mpDoAdvanceFrame();
	}
	if (framerate == 0)
	    tv.tv_sec = tv.tv_usec = 0;
	doNextFrame = mpFetchCommandOrData(playing ? &tv : NULL, 0);
	if (stream->film_has_ended && !loopFlag)
	    playing = 0;
	while (mpCmdQueueGet(&cmd, &param) == 0) {
	    switch (cmd) {
	      case MPEG_CMD_PLAY:
		playing = 1;
		doNextFrame = 1;
		/* Set up frame rate control variables */
		if (framerate < 0) {
		    perImage.tv_sec = 0;
		    perImage.tv_usec = 1000000 / stream->fps;
		} else if (framerate > 0) {
		    perImage.tv_sec = 0;
		    perImage.tv_usec = 1000000 / framerate;
		}
		break;
	      case MPEG_CMD_NEXT_FRAME:
		mpDoAdvanceFrame();
		break;
	      case MPEG_CMD_REWIND:
		mpDoRewind();
		break;
	      case MPEG_CMD_PAUSE:
		playing = 0;
		break;
	      case MPEG_CMD_REPAINT:
		mpDoRepaint();
		break;
	      case MPEG_CMD_PARENT_RESIZE:
		mpDoMove(0, 0);
		break;
	      case MPEG_CMD_SET_LOOP:
		loopFlag = param;
		break;
	      default:
		fprintf(stderr, "internal error in mpChildLoop\n");
	    }
	}
	if (doNextFrame && playing) {
	    /* timeout */
	    if (framerate)
		gettimeofday(&before, NULL);
	    mpDoAdvanceFrame();
	    if (stream->film_has_ended) {
		if (loopFlag)
		    mpDoRewind();
		else
		    playing = 0;
	    }
	    if (framerate) {
		gettimeofday(&after, NULL);
		used.tv_sec = after.tv_sec - before.tv_sec;
		if ((used.tv_usec = after.tv_usec - before.tv_usec) < 0L) {
		    used.tv_usec += 1000000L;
		    --used.tv_sec;
		}
                tv.tv_sec = perImage.tv_sec - used.tv_sec;
		if ((tv.tv_usec = perImage.tv_usec - used.tv_usec) < 0L) {
		    tv.tv_usec += 1000000L;
		    --tv.tv_sec;
		}
		if (tv.tv_sec < 0)
                    tv.tv_sec = tv.tv_usec = 0;
	    }
	}
    }
}

static void
mpQueueCommandAndReturnStatus(int cmd, int parameter)
{
    if (mpCmdQueueAdd(cmd, parameter) < 0)
	mpClientReturnStatus(MPEG_ERR_OVERFLOW);
    else
	mpClientReturnStatus(MPEG_OK);
}



/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

/* if (!waitForData) returns: 1 if timeout, 0 otherwise */
/* if (waitForData)  returns: -1 if read error, # bytes read otherwise
 *                                              0 in case of EOS */
int
mpFetchCommandOrData(struct timeval *tv, int waitForData)
{
    int cmd, done = 0, sel, ret = 0;
    struct fd_set fdset;

    if (waitForData && endOfStream)
	return 0;
    while (!done) {
	FD_ZERO(&fdset);
	FD_SET(pipeCmd, &fdset);
	sel = xselect(FD_SETSIZE, &fdset, NULL, NULL, tv);
	if ((ret = (sel == 0 
		    || (tv && tv->tv_sec == 0 && tv->tv_usec == 0))) != 0)
	    done = 1;
	if (sel > 0 && FD_ISSET(pipeCmd, &fdset)) {
	    /* file descriptor ready for reading */
	    cmd = mpClientReceiveCommand();
	    switch (cmd) {
	      case -1:
		/* error message already given by mpClientReceiveCommand */
		exit(0);
	      case MPEG_CMD_PLAY:
	      case MPEG_CMD_PAUSE:
	      case MPEG_CMD_REPAINT:
	      case MPEG_CMD_PARENT_RESIZE:
		mpQueueCommandAndReturnStatus(cmd, 0);
		if (!waitForData)
		    done = 1;
		break;
	      case MPEG_CMD_REWIND:
		if (!mpStreamIsSeekable())
		    mpClientReturnStatus(MPEG_ERR_NOT_SEEKABLE);
		else
		    mpQueueCommandAndReturnStatus(cmd, 0);
		if (!waitForData)
		    done = 1;
		break;
	      case MPEG_CMD_NEXT_FRAME:
		if (!stream)
		    mpClientReturnStatus(MPEG_ERR_NO_MOVIE);
		else if (stream->film_has_ended) {
		    /* possibly rewind here */
		    mpClientReturnStatus(MPEG_FILM_HAS_ENDED);
		} else
		    mpQueueCommandAndReturnStatus(cmd, 0);
		if (!waitForData)
		    done = 1;
		break;
	      case MPEG_CMD_QUIT:
		mpClientReturnStatus(MPEG_OK);
		mpCommonFinish();
		exit(0);
		break;
	      case MPEG_CMD_SET_LOOP: {
		  int onoff;

		  onoff = mpClientReceiveIntParameter();
		  if (onoff && !mpStreamIsSeekable())
		      mpClientReturnStatus(MPEG_ERR_NOT_SEEKABLE);
		  else
		      mpQueueCommandAndReturnStatus(cmd, onoff);
	        }
		if (!waitForData)
		    done = 1;
		break;
	      case MPEG_CMD_QUERY_WANTED_BYTES:
		mpClientReturnStatus(MPEG_OK);
		mpClientReturnInt(mpfQueryWantedBytes(stream->xfile));
		break;
	      case MPEG_CMD_SEND_BYTES: {
		  int n, wanted;

		  mpClientReturnStatus(MPEG_OK);
		  n = mpClientReceiveIntParameter();
		  wanted = mpfQueryWantedBytes(stream->xfile);
		  if (wanted < n)
		      n = wanted;
		  mpClientReturnInt(n);
		  ret = mpfGetDataFromFD(stream->xfile, pipeCmd, n);
		  mpClientReturnStatus(MPEG_OK);
	        }
     	        if (waitForData)
		    done = 1;
	        break;
	      case MPEG_CMD_END_OF_STREAM:
		mpClientReturnStatus(MPEG_OK);
		endOfStream = 1;
		ret = 0;
		if (waitForData)
		    done = 1;
		break;
	      default:
		fprintf(stderr, "video child received unknown command.\n");
		mpClientReturnStatus(MPEG_ERR_UNKNOWN);
	    }
	}
    }
    return ret;
}

void
mpDoMove(unsigned w, unsigned h)
{
    int px, py;
    int x, y;
    unsigned pw, ph, pbord, pdepth;
    unsigned bord, depth;
    Window proot, root;

    if (!w)
	XGetGeometry(xinfo.display, xinfo.window,
		     &root, &x, &y, &w, &h, &bord, &depth);
    XGetGeometry(xinfo.display, xinfo.parent,
		 &proot, &px, &py, &pw, &ph, &pbord, &pdepth);
    if ((x = ((int) pw - w) / 2) < 0)
	x = 0;
    if ((y = ((int) ph - h) / 2) < 0)
	y = 0;
    XMoveResizeWindow(xinfo.display, xinfo.window, x, y, w, h);
    XFlush(xinfo.display);
}

void
mpClientReturnStatus(int status)
{
    char st;

    st = status;
    if (write(pipeStat, &st, 1) != 1) {
	perror("unable to talk to parent process");
	exit(1);
    }
}

int
mpServerReceiveStatus(MpegStream stream)
{
    char st;

    if (!stream || read(stream->pipeStat, &st, 1) != 1) {
	perror("unable to read from child process");
	return MPEG_ERR_CHILD;
    }
    return st;
}

void
mpClientReturnInt(int value)
{
    if (write(pipeStat, &value, sizeof(int)) != sizeof(int)) {
	perror("unable to talk to parent process");
	exit(1);
    }
}

int
mpServerReceiveInt(MpegStream stream)
{
    int val;

    if (!stream || read(stream->pipeStat, &val, sizeof(int)) != sizeof(int)) {
	perror("unable to read from child process");
	return MPEG_ERR_CHILD;
    }
    return val;
}

int
mpServerSendCommand(MpegStream stream, int command)
{
    char cmd;

    cmd = command;
    if (!stream || write(stream->pipeCmd, &cmd, 1) != 1) {
	perror("unable to talk to child process");
	return MPEG_ERR_CHILD;
    }
    return mpServerReceiveStatus(stream);
}

int
mpServerSendIntCommand(MpegStream stream, int command, int parameter)
{
    char cmd;

    cmd = command;
    if (!stream 
	|| write(stream->pipeCmd, &cmd, 1) != 1
	|| write(stream->pipeCmd, &parameter, sizeof(int)) != sizeof(int)) {
	perror("unable to talk to child process");
	return MPEG_ERR_CHILD;
    }
    return mpServerReceiveStatus(stream);
}

int
mpClientReceiveCommand(void)
{
    char cmd;

    if (read(pipeCmd, &cmd, 1) != 1) {
	perror("unable to read from parent process");
	exit(1);
    }
    return cmd;
}

int
mpClientReceiveIntParameter(void)
{
    int parameter;

    if (read(pipeCmd, &parameter, sizeof(int)) != sizeof(int)) {
	perror("unable to read from parent process");
	exit(1);
    }
    return parameter;
}

MpegStream
mpNewStream(Display **dpy)
{
    MpegStream stream = NULL;
    pid_t pid;
    int pipeCommand[2], pipeStatus[2];

    if (pipe(pipeCommand) < 0 || pipe(pipeStatus) < 0) {
	perror("pipe");
	return NULL;
    }
    if ((pid = fork()) == 0) {
	/* child */
	struct sigaction sact;

	mpThisIsTheChild = 1;

	sigemptyset(&sact.sa_mask);
	sact.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &sact, NULL);

	close(ConnectionNumber(*dpy));
	/* TODO: should open same display as dpy */
	if ((*dpy = XOpenDisplay(NULL)) == NULL) {
	    fprintf(stderr, "unable to open default display\n");
	    exit(1);
	}
	pipeCmd = pipeCommand[0];
	close(pipeCommand[1]);
	pipeStat = pipeStatus[1];
	close(pipeStatus[0]);
	mpClientReturnStatus(127);
/* TODO: remove this sleep! It's here to make sure the
 * child doesn't write to a non-existing parent window. */
sleep(1);
    } else if (pid > 0) {
	/* parent */
	struct sigaction sact;

	if ((stream = malloc(sizeof(struct MpegStreamContainer))) == NULL) {
	    fprintf(stderr, "out of memory\n");
	    return NULL;
	}

	sigemptyset(&sact.sa_mask);
	sact.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &sact, NULL);

	stream->child = pid;
	stream->pipeCmd = pipeCommand[1];
	close(pipeCommand[0]);
	stream->pipeStat = pipeStatus[0];
	close(pipeStatus[1]);
	if (mpServerReceiveStatus(stream) != 127) {
	    fprintf(stderr, "child born dead\n");
	    free(stream);
	    return NULL;
	}
    } else {
	/* fork error */
	perror("fork");
	return NULL;
    }
    return stream;
}

void
mpChildNewStreamFromFile(const char *filename,
			 Display *display,
			 Window parent,
			 Window applicationMainWindow,
			 DitherType dither,
			 int frameRate,
			 int loop,
			 double gammaCorrection,
			 double chromaCorrection,
			 int useSharedMemory,
			 int useOwnColorMap)
{
    FILE *input;

    if ((input = fopen(filename, "rb")) == NULL) {
	fprintf(stderr, "could not open `%s' for reading\n", filename);
	exit(1);
    }
    mpCommonInit(display, parent, applicationMainWindow,
		 &dither, frameRate, loop,
		 gammaCorrection, chromaCorrection,
		 useSharedMemory, useOwnColorMap);
    stream = NewVidStream((unsigned int) BUF_LENGTH_FILE);
    stream->input = input;
    stream->xfile = NULL;

    mpChildLoop(); /* never returns */
}

void
mpChildNewStreamPushed(Display *display,
		       Window parent,
		       Window applicationMainWindow,
		       DitherType dither,
		       int frameRate,
		       double gammaCorrection,
		       double chromaCorrection,
		       int useSharedMemory,
		       int useOwnColorMap)
{
    XFILE *xf;

    if ((xf = mpfNewXFile(BUF_LENGTH_XFILE)) == NULL) {
	fprintf(stderr, "could not create XFILE\n");
	exit(1);
    }
    mpCommonInit(display, parent, applicationMainWindow,
		 &dither, frameRate, 0,
		 gammaCorrection, chromaCorrection,
		 useSharedMemory, useOwnColorMap);
    stream = NewVidStream((unsigned int) BUF_LENGTH_PUSHED);
    stream->xfile = xf;
    stream->input = NULL;

    mpChildLoop(); /* never returns */
}
