/* $Id$ */
/**************************************************************************
 *
 *  FILE            xtra.c
 *  MODULE OF       mpeg_play library.
 *
 *  DESCRIPTION     Interface to X11-routines.
 *
 *  WRITTEN BY      Sverre H. Huseby <sverrehu@ifi.uio.no>
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <values.h>
#include <string.h>

#include <X11/Xlib.h>

#include "xtra.h"

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

static int errBadValue;
static int (*oldBadValueHandler)(Display *, XErrorEvent *);


/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static int
xBadValueHandler(Display *dpy, XErrorEvent *err)
{
    if (err->error_code == BadValue)
	errBadValue = 1;
    else if (oldBadValueHandler)
	oldBadValueHandler(dpy, err);
    return 0;
}


/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

Status
xAddWMColormapWindow(Display *dpy, Window mainWin, Window colWin)
{
    Window *oldList, *newList;
    int    n, q, mustfree = 0;
    Status ret;

    if (XGetWMColormapWindows(dpy, mainWin, &oldList, &n)) {
	newList = malloc((n + 1) * sizeof(Window));
	newList[0] = colWin;
	for (q = 0; q < n; q++)
	    newList[q + 1] = oldList[q];
	mustfree = 1;
    } else {
	newList = &colWin;
	n = 0;
    }
    ret = XSetWMColormapWindows(dpy, mainWin, newList, n + 1);
    if (mustfree) {
	free(newList);
	XFree(oldList);
    }
    return ret;
}

Status
xRemoveWMColormapWindow(Display *dpy, Window mainWin, Window colWin)
{
    Window *oldList, *newList;
    int    n, q, w;
    Status ret;

    if (!XGetWMColormapWindows(dpy, mainWin, &oldList, &n))
	return 0;
    newList = malloc(n * sizeof(Window));
    for (q = w = 0; q < n; q++)
	if (oldList[q] != colWin)
	    newList[w++] = oldList[q];
    ret = XSetWMColormapWindows(dpy, mainWin, newList, w);
    free(newList);
    XFree(oldList);
    return ret;
}

Window
xGetParentWindow(Display *dpy, Window w)
{
    Window root, parent, *children;
    unsigned nchildren;

    if (XQueryTree(dpy, w, &root, &parent, &children, &nchildren) != 0) {
	XFree(children);
	return parent;
    }
    return 0;
}

Colormap
xGetParentColormap(Display *dpy, Window w)
{
    XWindowAttributes wa;
    Colormap ret;

    do {
        if ((w = xGetParentWindow(dpy, w)) == 0
	    || XGetWindowAttributes(dpy, w, &wa) == 0) {
            ret = XDefaultColormap(dpy, DefaultScreen(dpy));
	    break;
	}
	if (wa.colormap != None) {
	    ret = wa.colormap;
	    break;
	}
    } while (w);
    return ret;
}

Status
xAllocClosestColor(Display *dpy, Colormap cm, XColor *xc)
{
#define MAX_COL 512
#define KEEP
    int    q;
    Status ret;
#ifdef KEEP
    static int      nCol = 0;
    static XColor   *color = NULL;
    static double   *dist = NULL;
    static Colormap lastMap = 0;
#else
    int    nCol;
    XColor *color;
    double *dist;
#endif
    double rdiff, gdiff, bdiff, best;
    int    bestIdx;

    /* first try to do a normal allocation */
    if ((ret = XAllocColor(dpy, cm, xc)) != 0)
	return ret; /* ok, that worked! */

    /* that didn't work. */
    /* try allocating the closest among the first MAX_COL available colors */
#ifdef KEEP
    if (!color && (color = malloc(MAX_COL * sizeof(XColor))) == NULL)
	return 0;
    if (!dist && (dist = malloc(nCol * sizeof(double))) == NULL)
	return 0;
#else
    if ((color = malloc(MAX_COL * sizeof(XColor))) == NULL)
	return 0;
    if ((dist = malloc(nCol * sizeof(double))) == NULL) {
	free(color);
	return 0;
    }
#endif

    /* first fetch the colors */
#ifdef KEEP
    if (cm != lastMap) {
	lastMap = cm;
#endif
	errBadValue = 0;
	oldBadValueHandler = XSetErrorHandler(xBadValueHandler);
	for (nCol = 0; nCol < MAX_COL; nCol++) {
	    color[nCol].pixel = nCol;
	    XQueryColor(dpy, cm, &color[nCol]);
	    if (errBadValue)
		break;
	}
	XSetErrorHandler(oldBadValueHandler);
#ifdef KEEP
    }
#endif

    /* build table of `distances' */
    for (q = 0; q < nCol; q++) {
	rdiff = xc->red - color[q].red;
	rdiff *= rdiff;
	gdiff = xc->green - color[q].green;
	gdiff *= gdiff;
	bdiff = xc->blue - color[q].blue;
	bdiff *= bdiff;
	dist[q] = rdiff + gdiff + bdiff;
    }
    /* then find the closest one, and allocate it. */
    /* it may be that the alloc fails. in that case, find the next closest */
    do {
	best = MAXDOUBLE;
	bestIdx = -1;
	for (q = 0; q < nCol; q++)
	    if (dist[q] < best) {
		best = dist[q];
		bestIdx = q;
	    }
	if (best < MAXDOUBLE) {
	    /* try to allocate */
	    if ((ret = XAllocColor(dpy, cm, &color[bestIdx])) != 0) {
		memcpy(xc, &color[bestIdx], sizeof(XColor));
		goto finish;
	    } else {
		dist[bestIdx] = MAXDOUBLE; /* non-allocatable */
	    }
	}
    } while (best < MAXDOUBLE);

  finish:
#ifndef KEEP
    free(dist);
    free(color);
#endif

    return ret;
}
